use std::{
    collections::BTreeMap,
    sync::{Arc, RwLock},
};

use crate::state::SharedState;

use super::consumers::{Consumers, PartitionKey, Topic, TopicOffset};

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct StagingEvent {
    pub topic: String,
    pub key: String,
    pub id: Option<String>,
}

#[derive(Clone, Default)]
pub struct Staging {
    // Temporary until we've got an actual file disk store
    #[allow(clippy::complexity)]
    store: Arc<RwLock<BTreeMap<Topic, BTreeMap<PartitionKey, Vec<StagingEvent>>>>>,
}

impl Staging {
    pub fn new() -> Self {
        Self {
            store: Arc::default(),
        }
    }

    pub async fn publish(
        &self,
        staging_event: impl Into<StagingEvent>,
    ) -> anyhow::Result<TopicOffset> {
        let staging_event: StagingEvent = staging_event.into();
        let mut store = self.store.write().unwrap();
        tracing::trace!(
            topic = staging_event.topic,
            id = staging_event.id,
            "moving event to staging"
        );

        let offset = match store.get_mut(&staging_event.topic) {
            Some(part) => match part.get_mut(&staging_event.key) {
                Some(existing_key_part) => {
                    if staging_event.id.is_none()
                        || !existing_key_part.iter().any(|p| p.id == staging_event.id)
                    {
                        existing_key_part.push(staging_event);
                        existing_key_part.len() - 1
                    } else {
                        tracing::debug!(
                            topic = staging_event.topic,
                            id = staging_event.id,
                            "event already found, skipping"
                        );

                        existing_key_part.len() - 1
                    }
                }
                None => {
                    part.insert(staging_event.key.to_owned(), vec![staging_event]);
                    0
                }
            },
            None => {
                tracing::debug!(
                    topic = staging_event.topic,
                    id = staging_event.id,
                    "new topic, creating partition"
                );
                store.insert(
                    staging_event.topic.to_owned(),
                    BTreeMap::from([(staging_event.key.to_owned(), vec![staging_event])]),
                );

                0
            }
        };

        Ok(offset)
    }

    pub async fn get_topics(&self) -> anyhow::Result<Vec<String>> {
        let store = self.store.read().unwrap();

        Ok(store.keys().cloned().collect::<Vec<_>>())
    }

    pub async fn get_keys(&self, topic: impl Into<String>) -> anyhow::Result<Vec<String>> {
        let store = self.store.read().unwrap();

        let items = store
            .get(&topic.into())
            .map(|tree| tree.keys())
            .unwrap_or_default();

        Ok(items.cloned().collect::<Vec<_>>())
    }

    pub fn get_topic_offset(
        &self,
        topic: impl Into<String>,
        partition_key: Option<impl Into<PartitionKey>>,
        start: impl Into<TopicOffset>,
        end: impl Into<TopicOffset>,
    ) -> anyhow::Result<Vec<StagingEvent>> {
        let topic = topic.into();
        let partition_key = partition_key.map(|p| p.into()).unwrap_or_default();
        let start = start.into();
        let end = end.into();

        if start == end {
            return Ok(Vec::new());
        }

        if start > end {
            anyhow::bail!(
                "start cannot be greater than end, (start={}, end={})",
                start,
                end
            )
        }

        let store = self.store.read().unwrap();

        let partitions = match store.get(&topic) {
            Some(partitions) => partitions,
            None => {
                anyhow::bail!("topic doesn't exist in storage: {}", &topic);
            }
        };

        let partition = match partitions.get(&partition_key) {
            Some(partition) => partition,
            None => {
                anyhow::bail!(
                    "partition doesn't exist in storage, (topic={}, partition={})",
                    &topic,
                    &partition_key
                );
            }
        };

        if partition.len() < end {
            anyhow::bail!(
                "partition len is less than the offset, (partition_len={}, offset={})",
                partition.len(),
                end
            )
        }

        Ok(partition[start..=end].to_vec())
    }
}

pub trait StagingState {
    fn staging(&self) -> Staging;
}

impl StagingState for SharedState {
    fn staging(&self) -> Staging {
        self.staging.clone()
    }
}
